/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package es.rickyepoderi.maskedpwd;

import java.nio.charset.StandardCharsets;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.Provider;
import java.security.spec.InvalidKeySpecException;
import java.util.Arrays;
import java.util.stream.Collectors;
import org.wildfly.common.iteration.ByteIterator;
import org.wildfly.security.password.PasswordFactory;
import org.wildfly.security.password.WildFlyElytronPasswordProvider;
import org.wildfly.security.password.interfaces.MaskedPassword;
import org.wildfly.security.password.spec.EncryptablePasswordSpec;
import org.wildfly.security.password.spec.MaskedPasswordAlgorithmSpec;
import org.wildfly.security.password.spec.MaskedPasswordSpec;
import org.wildfly.security.password.util.ModularCrypt;

/**
 * <p>Little utility class to create masked passwords in order to be used inside
 * the <em>wildfly-config.xml</em> file. This is based on the work done in this
 * <a href="https://developer.jboss.org/people/aabdelsa/blog/2019/09/05/support-for-masked-passwords-in-the-client-xml-configuration">
 * blog entry</a>.</p>
 * 
 * @author rickyepoderi
 */
public class MaskedPasswordUtil {

    private static final Provider ELYTRON_PASSWORD_PROVIDER = new WildFlyElytronPasswordProvider();
    private static final String DEFAULT_ALGORITHM = "masked-MD5-DES";
    private static final String DEFAULT_KEY_MATERAL = "somearbitrarycrazystringthatdoesnotmatter";

    private String algorithm = DEFAULT_ALGORITHM;
    private String keyMaterial = DEFAULT_KEY_MATERAL;
    private int iterations = 0;
    private String salt = null;
    private String initializationVector = null;
    private String password = null;

    public static void usage(String error) {
        String nl = System.getProperty("line.separator");
        StringBuilder sb = new StringBuilder();
        if (error != null && !error.isEmpty()) {
            sb.append("ERROR: ").append(error).append(nl).append(nl);
        }
        sb.append("USAGE: java ").append(MaskedPasswordUtil.class.getSimpleName())
                .append(" OPTIONS password-to-mask").append(nl).append(nl);
        sb.append("OPTIONS:").append(nl);
        sb.append("  -a --algorithm alg: masked password algorithm (default: ").append(DEFAULT_ALGORITHM).append(")").append(nl);
        sb.append("  -s --salt salt: salt to use for the password (compulsory option - no default)").append(nl);
        sb.append("  -i --iter iterations: iteration count (compulsory option - no default)").append(nl);
        sb.append("  -k --keyMaterial material: key material (default: internal elytron value)").append(nl);
        sb.append("  -v --initializationVector vector: initialization vector (default: none)").append(nl);
        sb.append("  -h --help: shows this help").append(nl);
        sb.append(nl);
        throw new IllegalArgumentException(sb.toString());
    }

    public static String parseString(String[] args, int i) {
        if (i + 1 >= args.length) {
            usage(String.format("Option %s needs a value", args[i]));
        }
        return args[i + 1];
    }

    public static int parsePositiveInteger(String[] args, int i) {
        String value = parseString(args, i);
        try {
            int result = Integer.parseInt(value);
            if (result > 0) {
                return result;
            }
        } catch (NumberFormatException e) {
        }
        usage(String.format("Invalid positive number %s for option %s", value, args[i]));
        return 0;
    }

    public static String parseAlgorithm(String[] args, int i) {
        String alg = parseString(args, i);
        try {
            PasswordFactory.getInstance(alg, ELYTRON_PASSWORD_PROVIDER);
        } catch (NoSuchAlgorithmException e) {
            usage(String.format("Invalid algorithm %s for option %s (allowed values %s)", alg, args[i],
                    ELYTRON_PASSWORD_PROVIDER.getServices().stream().map(s -> s.getAlgorithm())
                            .filter(a -> a.startsWith("masked-")).collect(Collectors.toSet())));
        }
        return alg;
    }

    public void parseArguments(String... args) {
        for (int i = 0; i < args.length; i++) {
            switch (args[i]) {
                case "-a":
                case "--algorithm":
                    this.algorithm = parseAlgorithm(args, i++);
                    break;
                case "-s":
                case "--salt":
                    this.salt = parseString(args, i++);
                    break;
                case "-i":
                case "--iter":
                    this.iterations = parsePositiveInteger(args, i++);
                    break;
                case "-k":
                case "--keyMaterial":
                    this.keyMaterial = parseString(args, i++);
                    break;
                case "-v":
                case "--initializationVector":
                    this.initializationVector = parseString(args, i++);
                    break;
                case "-h":
                case "--help":
                    usage(null);
                    break;
                default:
                    // no more arguments => password to mask
                    if (i < args.length - 1) {
                        usage("Invalid number of arguments");
                    }
                    this.password = args[i];
            }
        }
        if (this.password == null) {
            usage("No password passed for masking");
        }
        if (this.salt == null) {
            usage("Option for salt (-s or --salt) is compulsory");
        }
        if (this.iterations == 0) {
            usage("Option for iterations (-i or --iter) is compulsory");
        }
    }

    public void execute() throws InvalidKeySpecException, InvalidKeyException, NoSuchAlgorithmException {
        PasswordFactory passwordFactory = PasswordFactory.getInstance(algorithm, ELYTRON_PASSWORD_PROVIDER);

        MaskedPasswordAlgorithmSpec maskedAlgorithmSpec = new MaskedPasswordAlgorithmSpec(keyMaterial.toCharArray(),
                iterations, salt.getBytes(StandardCharsets.UTF_8),
                initializationVector == null? null : initializationVector.getBytes(StandardCharsets.UTF_8));
        EncryptablePasswordSpec encryptableSpec = new EncryptablePasswordSpec(password.toCharArray(), maskedAlgorithmSpec);

        MaskedPassword maskedPassword = (MaskedPassword) passwordFactory.generatePassword(encryptableSpec);

        // print the data and snippet
        System.out.println("Masked Password: " + ModularCrypt.encodeAsString(maskedPassword));
        StringBuilder sb = new StringBuilder("<masked-password");
        if (!DEFAULT_ALGORITHM.equals(maskedPassword.getAlgorithm())) {
            sb.append(" algorithm=\"").append(maskedPassword.getAlgorithm()).append("\"");
        }
        if (!Arrays.equals(maskedPassword.getInitialKeyMaterial(), DEFAULT_KEY_MATERAL.toCharArray())) {
            sb.append(" key-material=\"").append(maskedPassword.getInitialKeyMaterial()).append("\"");
        }
        sb.append(" iteration-count=\"").append(maskedPassword.getIterationCount()).append("\"");
        sb.append(" salt=\"").append(ByteIterator.ofBytes(maskedPassword.getSalt()).asUtf8String().drainToString()).append("\"");
        if (maskedPassword.getInitializationVector() != null) {
            sb.append(" initialization-vector=\"").append(ByteIterator.ofBytes(maskedPassword.getInitializationVector()).base64Encode().drainToString()).append("\"");
        }
        sb.append(" masked-password=\"").append(ByteIterator.ofBytes(maskedPassword.getMaskedPasswordBytes()).base64Encode().drainToString()).append("\"/>");
        System.out.println("Snippet: " + sb.toString());

        // verify the masked password is the encryption of the clear password
        MaskedPasswordSpec maskedPasswordSpec = new MaskedPasswordSpec(maskedPassword.getInitialKeyMaterial(),
                maskedPassword.getIterationCount(), maskedPassword.getSalt(),
                maskedPassword.getMaskedPasswordBytes(), maskedPassword.getInitializationVector());

        MaskedPassword restored = (MaskedPassword) passwordFactory.generatePassword(maskedPasswordSpec);
        System.out.println(String.format("Password Verified '%b'", passwordFactory.verify(restored, password.toCharArray())));
    }

    public static void main(String... args) throws Exception {
        MaskedPasswordUtil util = new MaskedPasswordUtil();
        util.parseArguments(args);
        util.execute();
    }
}
